@@ -420,12 +420,12 @@ int common_block_args_process(int argc, char *argv[],
 	return 0;
 }
 
-int debugfs_read_inode_full(ext2_ino_t ino, struct ext2_inode * inode,
-			const char *cmd, int bufsize)
+int debugfs_read_inode2(ext2_ino_t ino, struct ext2_inode * inode,
+			const char *cmd, int bufsize, int flags)
 {
 	int retval;
 
-	retval = ext2fs_read_inode_full(current_fs, ino, inode, bufsize);
+	retval = ext2fs_read_inode2(current_fs, ino, inode, bufsize, flags);
 	if (retval) {
 		com_err(cmd, retval, "while reading inode %u", ino);
 		return 1;
@@ -446,15 +446,14 @@ int debugfs_read_inode(ext2_ino_t ino, struct ext2_inode * inode,
 	return 0;
 }
 
-int debugfs_write_inode_full(ext2_ino_t ino,
-			     struct ext2_inode *inode,
-			     const char *cmd,
-			     int bufsize)
+int debugfs_write_inode2(ext2_ino_t ino,
+			 struct ext2_inode *inode,
+			 const char *cmd,
+			 int bufsize, int flags)
 {
 	int retval;
 
-	retval = ext2fs_write_inode_full(current_fs, ino,
-					 inode, bufsize);
+	retval = ext2fs_write_inode2(current_fs, ino, inode, bufsize, flags);
 	if (retval) {
 		com_err(cmd, retval, "while writing inode %u", ino);
 		return 1;
