@@ -91,6 +91,7 @@ void proceed_question(int delay)
 {
 	char buf[256];
 	const char *short_yes = _("yY");
+	const char *english_yes = "yY";
 
 	fflush(stdout);
 	fflush(stderr);
@@ -101,14 +102,16 @@ void proceed_question(int delay)
 			return;
 		}
 		signal(SIGALRM, alarm_signal);
-		printf(_("Proceed anyway (or wait %d seconds) ? (y,n) "),
+		printf(_("Proceed anyway (or wait %d seconds to proceed) ? (y,N) "),
 		       delay);
 		alarm(delay);
 	} else
-		fputs(_("Proceed anyway? (y,n) "), stdout);
+		fputs(_("Proceed anyway? (y,N) "), stdout);
 	buf[0] = 0;
 	if (!fgets(buf, sizeof(buf), stdin) ||
-	    strchr(short_yes, buf[0]) == 0) {
+	    strchr(_("nN"), buf[0]) ||
+	    !(strchr(short_yes, buf[0]) ||
+	      strchr(english_yes, buf[0]))) {
 		putc('\n', stdout);
 		exit(1);
 	}
